from qubtree import Node
import QUB
from Dialogs import *
from math import *
from string import *
import re


fitters = {} # dictionary[name] -> factory (typically the type object of the fitter)

# To add your own curve types, subclass NLFitter in a separate file.
# See GaussFit.py for an example.

# At minimum, you must provide these methods:
#     __init__:  call initParams
#    func:       calc. func value of one component, with params in the list paramvals, evaluated at x

# It's also a good idea to provide
#    deriv:      calc. partial derivative w.r.t. each param for one component, evaluated at x

# For user convenience, you can adjust the parameters when:
#    * a component is added:         initComponent()
#    * data is assigned or changed:  setParamsFromData()
#    * fit() finishes:               onFitDone()

# You can expose additional options in the user interface
# (e.g. choice of fitting algorithms?)
# by overriding promptOptions() -- see commented-out example in GaussFit.py.

MARQUARDT = 0
SIMPLEX   = 1
BFGS      = 2
SIMANNEAL = 3

def componentToList(comp):
    lst = []
    param = comp.child
    while not (param.isNull or (param.data.count == 0)):
        lst.append( param.data[0] )
        param = param.sibling
    return lst

class NLFitter:
    def __init__(self):
        self.state = Node()
        self.paramsNode = self.state['Params']
        self.paramlists = []
        self.ncomponent = 0
        self.data = Node('FitData')
        X = self.data['X']
        X.data = 1.0
        self.X = X.data
        Y = self.data['Y']
        Y.data = 1.0
        self.Y = Y.data
        self.N = 1
        self.algo = MARQUARDT

    def initParams(self, nparamToFit, *args):                # subclasses call initParams in their __init__
        param = self.paramsNode['Param']
        while not param.child.isNull:
            param.remove( param.child )
        self.paramlists = []
            
        self.nparam = 0
        for p in args:
            param[p[0]].data = p[1:]
            self.nparam = self.nparam + 1
        self.params = [param]
        self.ncomponent = 1
        self.nparamToFit = nparamToFit

    def initComponent(self, param):
        pass # use values from component 0

    def setParamsFromData(self):
        pass # unchanged

    def func(self, paramvals, x):
        return 0.0

    def deriv(self, paramvals, x):
        return []

    def onFitDone(self):
        pass

    def getEqn(self):
        return ''

    def promptOptions(self):
        # QUB.Dialog(message, buttons, fields) -> { field : value, ... }
        # fields = [ QUBB.Dialog*(name, ...) ]

        chosenalgo = ''
        if self.algo == MARQUARDT:
            chosenalgo = 'Marquardt'
        elif self.algo == SIMPLEX:
            chosenalgo = 'Simplex'
        elif self.algo == BFGS:
            chosenalgo = 'BFGS'
        else:
            chosenalgo = 'Simulated Annealing'
            
        options = QUB.Dialog('Curve fitting options',
                             ['OK', 'Cancel'],
                             [ DialogRadioGroup('algo', 'Algorithm:',
                                           ["Marquardt", "Simplex", "BFGS", "Simulated Annealing"],
                                                chosenalgo) ])
        if options['button'] == 'OK':
            algo = options['algo']
            if algo == 'Marquardt':
                self.algo = MARQUARDT
            elif algo == 'Simplex':
                self.algo = SIMPLEX
            elif algo == 'BFGS':
                self.algo = BFGS
            else: # simulated annealing
                self.algo = SIMANNEAL

    def fit(self, maxIter=10, precision=0.01):
        err = QUB.NLRegFit(self, maxIter, precision)
        
        self.onFitDone()
        
        if err == -1:
            QUB.Report('Fit error: singular hessian matrix')
        elif err == -2:
            QUB.Report("Fit error: too high Marquardt's parameter")
        elif err == -3:
            QUB.Report("Fit error: non-convergence")
        
    def setNumComponents(self, n):
        while n < len(self.params):
            self.paramsNode.remove( self.params.pop() )
        while n > len(self.params):
            newparam = self.params[0].clone()
            self.paramsNode.append( newparam )
            self.params.append( newparam )
            self.initComponent( newparam )
        self.ncomponent = n

    def getNumComponents(self):
        return self.ncomponent

    def getParamList(self, component_i):
        while len(self.paramlists) <= component_i:
            newlist = []
            param = self.params[ len(self.paramlists) ].child
            while not param.isNull:
                newlist.append( param.data[0] )
                param = param.sibling
            self.paramlists.append( newlist )
            
        return self.paramlists[ component_i ]

    def updateParamLists(self):
        for i in range(self.ncomponent):
            lst = self.getParamList(i)
            comp = self.params[i]
            param = comp.child
            for j in range(self.nparam):
                lst[j] = param.data[0]
                param = param.sibling

    def useData(self, dataset):
        self.data = dataset
        self.X = dataset['X'].data
        self.Y = dataset['Y'].data
        self.N = min(self.X.count, self.Y.count)
        self.setParamsFromData()
        
    def getCurves(self, curves):
        X = self.X
        Y = self.Y
        N = self.N
    
        curveList = []
        curve = curves.child
        while not curve.isNull:
            if curve.data.count != N:
                curve.data = Y
            curveList.append(curve)
            curve = curve.sibling

        nc = self.ncomponent
        while len(curveList) <= nc:
            curve = Node('Component')
            curve.data = Y
            curves.append(curve)
            curveList.append(curve)

        sumCurve = curveList[0].data
        for i in range(N):
            sumCurve[i] = 0.0
        
        for comp in range(self.ncomponent):
            curve = curveList[comp+1].data
            paramvals = componentToList(self.params[comp])
            for i in range(N):
                curve[i] = self.func(paramvals, X[i])
                sumCurve[i] = sumCurve[i] + curve[i]

    def getN(self):
        return self.N
    def getNComponent(self):
        return self.ncomponent
    def getNParam(self):
        return self.nparam


def MissingName(msg): # extract the unknown name from a NameError message
    m = re.search("'([^']+)'", msg)
    if not m:
        raise NameError, msg
    return m.group(1)    


class CustomFitter(NLFitter):
    def __init__(self):
        NLFitter.__init__(self)
        self.initParams( 3,
                         ('Amp', 1.0, 1.0e-20, 1.0e20),
                         ('Mean', 1.0, -1.0e20, 1.0e20),
                         ('Std', 1.0, 1.0e-20, 1.0e20) )
        self.eqn = 'Amp * exp( -(x - Mean) * (x - Mean) / (2 * Std * Std) )'
        self.customFunc = lambda x, Amp, Mean, Std: Amp * exp( -(x - Mean) * (x - Mean) / (2 * Std * Std) )
        self.customDerivs = {'Amp' : None, 'Mean' : None, 'Std' : None}
        self.customDerivStrs = {'Amp' : '', 'Mean' : '', 'Std' : ''}
        
    def getEqn(self):
        return self.eqn

    def func(self, paramvals, x):
        try:
            return apply( self.customFunc, [x] + paramvals )
        except:
            return 0.0

    def deriv(self, paramvals, x):
        args = [x] + paramvals
        if len(self.customDerivs) == len(paramvals):
            onecomp = self.params[0]
            return map( lambda param_i: apply( self.customDerivs[onecomp[param_i].name], args ),
                        range(len(paramvals)) )
        else:
            return []

    def promptOptions(self):
        msg = 'Please enter an expression in Python style'
        neweqn = self.eqn
        valid = 0

        chosenalgo = ''
        if self.algo == MARQUARDT:
            chosenalgo = 'Marquardt'
        elif self.algo == SIMPLEX:
            chosenalgo = 'Simplex'
        elif self.algo == BFGS:
            chosenalgo = 'BFGS'
        else:
            chosenalgo = 'Simulated Annealing'
            
        while not valid:        
            options = QUB.Dialog(msg,
                                 ['OK', 'Cancel'],
                                 [ DialogText('eqn', 'y = f(x) = ', neweqn),
                                   DialogRadioGroup('algo', 'Algorithm:',
                                           ["Marquardt", "Simplex", "BFGS", "Simulated Annealing"],
                                                    chosenalgo) ])
            if options['button'] == 'Cancel':
                return
            else:
                chosenalgo = options['algo']
                try:
                    neweqn = options['eqn']
                    self.setCustomEqn( neweqn )
                    valid = 1
                except Exception, x:
                    msg = str(x)

        if chosenalgo == 'Marquardt':
            self.algo = MARQUARDT
        elif chosenalgo == 'Simplex':
            self.algo = SIMPLEX
        elif chosenalgo == 'BFGS':
            self.algo = BFGS
        else: # simulated annealing
            self.algo = SIMANNEAL

        msg = 'Derivatives are optional, but can improve the fit'
        valid = 0
        params = map(lambda i: self.params[0][i].name, range(len(self.params[0])))
        while not valid:
            options = QUB.Dialog(msg,
                                 ['OK', 'No Derivatives'],
                                 map(lambda name: DialogText(name, 'dy/d_'+name, self.customDerivStr(name)), params))
            if options['button'] == 'No Derivatives':
                self.customDerivs = {}
                valid = 1
            else:
                try:
                    newderivs = map( lambda name: options[name], params )
                    self.setCustomDerivs( params, newderivs )
                    valid = 1
                except Exception, x:
                    msg = str(x)

    def setCustomEqn(self, neweqn):
        args = ['x']
        valid = 0
        newfunc = lambda : 0.0

        while not valid:
            try:
                newfunc = eval( 'lambda ' + join(args, ',') + ': ' + neweqn )
                fval = float( apply( newfunc, range(len(args)) ) )
                valid = 1
            except NameError, msg:
                args.append( MissingName(str(msg)) )

        paramInits = [len(args) - 1]
        for arg in args[1:]:
            paramInits.append( (arg, 1.0, 1.0e-20, 1.0e20) )
        apply(self.initParams, paramInits)
        self.eqn = neweqn
        self.customFunc = newfunc

        for pname in self.customDerivs.keys():
            if not pname in args:
                del self.customDerivs[pname]
                del self.customDerivStrs[pname]

    def customDerivStr(self, pname):
        try:
            return self.customDerivStrs[pname]
        except:
            return ''
    
    def setCustomDerivs(self, paramnames, newderivstrs):
        args = ['x'] + paramnames
        lambdaargs = 'lambda ' + join(args, ',') + ': '
        newderivs = {}

        for i in range( len(newderivstrs) ):
            self.customDerivStrs[ paramnames[i] ] = newderivstrs[i]

        for i in range( len(newderivstrs) ):
            newderiv = eval( lambdaargs + newderivstrs[i] )
            dval = float( apply( newderiv, range(len(args)) ) )
            newderivs[ paramnames[i] ] = newderiv

        self.customDerivs = newderivs

fitters['Custom'] = CustomFitter


# >>> fitter = FitterType()

# >>> fitter.setNumComponents(1)
# >>> print fitter.getNumComponents()
# 1

# >>> print fitter.paramsNode
# Params
# {
#     Param
#     {
#         a = 1.2
#         b = 3.4
#     }
#     ... # one Param node per component; subnodes should be in same order each Param
# }

# >>> dataset = Node('FitData')
# >>> dataset['X'].data = [1.2, 3, -2]
# >>> dataset['Y'].data = [0.3, -2, 5]
# >>> fitter.useData( dataset )        # ideally should update params to reasonable
# >>> print fitter.data                # values given data and numcomponents
# FitData
# {
#     X = 1.2   3   -2
#     Y = 0.3   -2  5
# }

# >>> for x in dataset['X'].data:
# ...     print fitter.calcY(0, x, nextX),    # y value of first component at x
#                                            # optional arg. nextX is X[i+1] if available
#                                            # else an extrapolation using ratio X[i]/X[i-1]
#                                            # (needed for LogExpoFit)

# >>> fitter.promptOptions()           # prompt user for extra options (or pass)

# >>> time2stop = 0
# >>> def iterCallback(continueFlag):
# ...     print "iter"
# ...     if time2stop:
# ...         continueFlag.data = 0
# ...     return continueFlag
# ... 
# >>> fitter.fit(maxiter=3, iterCallback, precision=0.01)
# iter
# iter
# iter

# first element is the sum of all the rest
def RebuildCurves(fitter, curves):
    data = fitter.data
    X = data['X'].data
    Y = data['Y'].data
    np = X.count
    
    curveList = []
    curve = curves.child
    while not curve.isNull:
        if curve.data.count != np:
            curve.data = Y
        curveList.append(curve)
        curve = curve.sibling

    nc = fitter.getNumComponents()
    while len(curveList) <= nc:
        curve = Node('Component')
        curve.data = Y
        curves.append(curve)
        curveList.append(curve)

    sumCurve = curveList[0].data
    for i in range(np):
        sumCurve[i] = 0.0
        
    for comp in range(nc):
        curve = curveList[comp+1].data
        for i in range(np-1):
            curve[i] = fitter.calcY(comp, X[i], X[i+1])
            sumCurve[i] = sumCurve[i] + curve[i]

        superLastX = X[np-1]
        try:
            superLastX = superLastX * (superLastX / X[np-2])
        except:
            pass

        curve[np-1] = fitter.calcY(comp, X[np-1], superLastX)
        sumCurve[np-1] = sumCurve[np-1] + curve[np-1]
        
